
require 'ipaddr'
require 'yaml'
require 'pp'

class IPAddr
  def next
    raise "next: Not supported for IPv6!" if self.ipv6?
    IPAddr::from_int(self.to_i + 1)
  end

  def IPAddr::from_int(i)
    s = [ (i & 0xff000000) >> 24, (i & 0xff0000) >> 16, (i & 0xff00) >> 8, i & 0xff].join('.')
    return IPAddr::new(s)
  end

  def +(i)
    if i > 1
      return self.next + (i-1)
    else
      return self.next
    end
  end

  def to_masked
    inspect.split(':')[2].split('>')[0].gsub('/',':')
  end
end

# a netgroup
class NetGroup
  attr_accessor :ul_latency, :ul_rate, :dl_latency, :dl_rate, :nbhosts, :systems, :children, :hosts, :links
  attr_reader :selfnbhosts
  
  def initialize(netmask = nil)
    @netmask = IPAddr::new(netmask) if netmask
    @children = []
    @hosts = []
    @nbhosts = 0
    @systems = []
    @links = []
  end

  attr_reader :netmask

  def netmask=(m)
    @netmask = IPAddr::new(m)
  end

  def ==(o)
    return @netmask == o.netmask
  end
  
  def include?(nm)
    return @netmask.include?(nm)
  end

  def add_group(g)
    added = false
    @children.each do |i|
      if i.include?(g.netmask)
        i.add_group(g)
        added = true
        break
      end
    end
    @children << g if not added
  end

  def add_link(l)
    added = false
    @children.each do |i|
      if i.include?(l.src)
        i.add_link(l)
        added = true
        break
      end
    end
    @links << l if not added
  end

  def add_host(h)
    added = false
    @children.each do |i|
      if i.include?(h.addr)
        i.add_host(h)
        added = true
        break
      end
    end
    @hosts << h if not added
  end

  def decorate(ul_latency = nil, dl_latency = nil, ul_rate = nil, dl_rate = nil, links = [])
    @ul_latency ||= ul_latency
    @dl_latency ||= dl_latency
    @ul_rate ||= ul_rate
    @dl_rate ||= dl_rate
    @links = @links + links
    nbh = 0
    @children.each { |n| nbh += n.decorate(@ul_latency,
                                           @dl_latency, @ul_rate, @dl_rate, @links) }
    @hosts.each { |h| h.decorate(@ul_latency,
                                           @dl_latency, @ul_rate, @dl_rate) }
    if nbh + @hosts.length > @nbhosts
      @nbhosts = nbh + @hosts.length 
      @selfnbhosts = 0
    else
      @selfnbhosts = @nbhosts - nbh
    end
    @nbhosts
  end

  def affect_hosts(hostpool, maxnb)
    # affects to own systems
    nbh = @selfnbhosts
    while nbh > 0
      raise "Not enough real systems ; missing at least for #{nbh} VN" if hostpool.empty?
      if nbh > maxnb
        @systems.push([hostpool.shift, maxnb])
        nbh -= maxnb
      else
        @systems.push([hostpool.shift, nbh])
        nbh = 0
      end
    end
    # affect to children
    @children.each { |c| c.affect_hosts(hostpool, maxnb) }
  end

  def write_commands(prefix)
    ip = @netmask.next
    totnbh = 0
      @systems.each do |t|
      s, n = t
      npipe = 1
      File::open(prefix + s, 'w') do |f|
        n.times do
          f.puts "/sbin/ifconfig bge0 #{ip} alias"
          ulrate = @ul_rate
          dlrate = @dl_rate
          ullat = @ul_latency
          dllat = @dl_latency
          @hosts.each do |h|
            if h.addr == ip
              ulrate = h.ul_rate
              dlrate = h.dl_rate
              ullat = h.ul_latency
              ullat = h.dl_latency
            end
          end
          ulrate = ulrate ? "bw #{ulrate}" : ""
          dlrate = dlrate ? "bw #{dlrate}" : ""
          ullat = ullat ? "delay #{ullat}" : ""
          dllat = dllat ? "delay #{dllat}" : ""
          f.puts "/sbin/ipfw -q add pipe #{npipe} ip from #{ip} to any out"
          f.puts "/sbin/ipfw -q pipe #{npipe} config #{ulrate || ""} #{ullat || ""}"
          f.puts "/sbin/ipfw -q add pipe #{npipe + 1} ip from any to #{ip} in"
          f.puts "/sbin/ipfw -q pipe #{npipe + 1} config #{dlrate || ""} #{dllat || ""}"
          npipe += 2
          ip = ip.next
        end
        @links.each do |l|
          rate = l.rate ? ("bw " + l.rate) : ""
          lat = l.latency ? ("delay " + l.latency) : ""
          f.puts "/sbin/ipfw -q add pipe #{npipe} ip from #{l.src.to_masked} to #{l.dst.to_masked}"
          f.puts "/sbin/ipfw -q pipe #{npipe} config #{rate} #{lat}"
          npipe += 1
        end
      end
    end
    @children.each { |c| c.write_commands(prefix) }
  end
end

# a tree of netgroups
class NetGroups
  def initialize(config)
    d = YAML::load(config)
    # read groups from config
    groups = []
    d['groups'].each { |dg|
      g = NetGroup::new
      raise "Not netmask specified!" if dg['netmask'].nil?
      g.netmask = dg['netmask']
      g.ul_latency = dg['ul_latency']
      g.dl_latency = dg['dl_latency']
      g.ul_rate = dg['ul_rate']
      g.dl_rate = dg['dl_rate']
      g.nbhosts = dg['nbhosts'] || 0
      groups << g
    }

    # sanity check: avoid dups
    for i in 0...groups.length
      for j in i+1...groups.length
        if groups[i] == groups[j]
          raise "Duplicate netmask #{groups[i].netmask}" 
        end
      end
    end
    @tree = NetGroup::new('0.0.0.0/0')
    # while the list of groups is not empty
    while not groups.empty?
      # find a root node (yes, O(n), but I don't care.)
      node = nil
      groups.each do |i|
        included = false
        groups.each do |j|
          if i != j and j.include?(i.netmask)
            included = true
            break
          end
        end
        if not included
          node = i
          break
        end
      end
      @tree.add_group(node)
      groups.delete(node)
    end
    # add hosts
    (d['hosts'] || []).each do |dh|
      h = Host::new
      raise 'Specifying an IP ADDR for an host is mandatory' if dh['addr'].nil?
      h.addr = dh['addr']
      h.ul_latency = dh['ul_latency']
      h.dl_latency = dh['dl_latency']
      h.ul_rate = dh['ul_rate']
      h.dl_rate = dh['dl_rate']
      @tree.add_host(h)
    end
    # add links
    (d['links'] || []).each do |dh|
      l = Link::new(dh['src'], dh['dst'])
      l.latency = dh['latency']
      l.rate = dh['rate']
      @tree.add_link(l)
      p dh
      if dh['two-way']
        l = Link::new(dh['dst'], dh['src'])
        l.latency = dh['latency']
        l.rate = dh['rate']
        @tree.add_link(l)
      end
    end
  end

  def decorate
    @tree.decorate
    self
  end

  def show
    pp @tree
    self
  end

  # affect some real hosts to some netgroups.
  # Rules :
  # 1) A real host can only contain hosts of the same netgroup
  # 2) A real host can contain maxnb hosts max.
  def affect_hosts(hostpool, maxnb)
    @tree.affect_hosts(hostpool, maxnb)
    self
  end

  def write_commands(prefix = 'cmds_')
    @tree.write_commands(prefix)
  end
end

class Host
  attr_accessor :ul_latency, :ul_rate, :dl_latency, :dl_rate

  attr_reader :addr

  def addr=(a)
    @addr = IPAddr::new(a)
  end

  def decorate(ul_latency = nil, dl_latency = nil, ul_rate = nil, dl_rate = nil)
    @ul_latency ||= ul_latency
    @dl_latency ||= dl_latency
    @ul_rate ||= ul_rate
    @dl_rate ||= dl_rate
  end
end

class Link
  attr_accessor :src, :dst, :latency, :rate
  def initialize(src, dst)
    @src = IPAddr::new(src)
    @dst = IPAddr::new(dst)
  end
end
